from functools import lru_cache

import luigi
import yaml

from servicecatalog_puppet import manifest_utils
from servicecatalog_puppet.workflow import tasks
from servicecatalog_puppet import constants


class ManifestMixen(object):
    manifest_file_path = luigi.Parameter()

    @property
    @lru_cache()
    def manifest(self):
        content = open(self.manifest_file_path, "r").read()
        return manifest_utils.Manifest(yaml.safe_load(content))


class SectionTask(tasks.PuppetTask, ManifestMixen):
    manifest_file_path = luigi.Parameter()
    puppet_account_id = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "manifest_file_path": self.manifest_file_path,
            "cache_invalidator": self.cache_invalidator,
        }

    @property
    @lru_cache()
    def manifest(self):
        content = open(self.manifest_file_path, "r").read()
        return manifest_utils.Manifest(yaml.safe_load(content))

    def handle_requirements_for(
        self,
        name,
        details,
        section_name_singular,
        section_name_plural,
        is_in_spoke_execution_mode,
        for_region_task_klass,
        for_account_task_klass,
        for_account_and_region_task_klass,
        task_klass,
        kwargs_to_use,
    ):
        if is_in_spoke_execution_mode:
            if section_name_plural == constants.LAUNCHES:
                if details.get("execution") != constants.EXECUTION_MODE_SPOKE:
                    return []
        else:
            if details.get("execution") == constants.EXECUTION_MODE_SPOKE:
                from servicecatalog_puppet.workflow import launch

                dependencies = list()
                for account_id in self.manifest.get_account_ids_used_for_section_item(
                    self.puppet_account_id, section_name_plural, name
                ):
                    dependencies.append(
                        launch.RunDeployInSpokeTask(
                            manifest_file_path=kwargs_to_use.get("manifest_file_path"),
                            puppet_account_id=self.puppet_account_id,
                            account_id=account_id,
                        )
                    )
                return dependencies

        dependencies = list()

        affinities_used = dict()
        is_a_dependency = False

        for manifest_section_name in constants.ALL_SECTION_NAMES:
            for n, d in self.manifest.get(manifest_section_name, {}).items():
                for dep in d.get("depends_on", []):
                    if (
                        dep.get("type") == section_name_singular
                        and dep.get("name") == name
                    ):
                        is_a_dependency = True
                        affinities_used[dep.get("affinity")] = True

        if is_a_dependency:
            if affinities_used.get(constants.AFFINITY_REGION):
                for region in self.manifest.get_regions_used_for_section_item(
                    self.puppet_account_id, section_name_plural, name
                ):
                    dependencies.append(
                        for_region_task_klass(**kwargs_to_use, region=region,)
                    )

            if affinities_used.get(constants.AFFINITY_ACCOUNT):
                for account_id in self.manifest.get_account_ids_used_for_section_item(
                    self.puppet_account_id, section_name_plural, name
                ):
                    dependencies.append(
                        for_account_task_klass(**kwargs_to_use, account_id=account_id,)
                    )

            if affinities_used.get(constants.AFFINITY_ACCOUNT_AND_REGION):
                for (
                    account_id,
                    regions,
                ) in self.manifest.get_account_ids_and_regions_used_for_section_item(
                    self.puppet_account_id, section_name_plural, name
                ).items():
                    for region in regions:
                        dependencies.append(
                            for_account_and_region_task_klass(
                                **kwargs_to_use, account_id=account_id, region=region,
                            )
                        )

            if affinities_used.get(section_name_singular):
                dependencies.append(task_klass(**kwargs_to_use))

        else:
            dependencies.append(task_klass(**kwargs_to_use))

        return dependencies



import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow import manifest as manifest_tasks
from servicecatalog_puppet.workflow import tasks as workflow_tasks, dependency


class CodeBuildRunBaseTask(workflow_tasks.PuppetTaskWithParameters):
    manifest_file_path = luigi.Parameter()

    @property
    def section_name(self):
        return constants.CODE_BUILD_RUNS


class ExecuteCodeBuildRunTask(
    CodeBuildRunBaseTask, manifest_tasks.ManifestMixen, dependency.DependenciesMixin
):
    code_build_run_name = luigi.Parameter()
    puppet_account_id = luigi.Parameter()

    region = luigi.Parameter()
    account_id = luigi.Parameter()

    ssm_param_inputs = luigi.ListParameter(default=[], significant=False)

    launch_parameters = luigi.DictParameter(default={}, significant=False)
    manifest_parameters = luigi.DictParameter(default={}, significant=False)
    account_parameters = luigi.DictParameter(default={}, significant=False)

    project_name = luigi.Parameter()
    requested_priority = luigi.IntParameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "code_build_run_name": self.code_build_run_name,
            "region": self.region,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        return dict(section_dependencies=self.get_section_dependencies())

    def run(self):
        yield DoExecuteCodeBuildRunTask(
            manifest_file_path=self.manifest_file_path,
            code_build_run_name=self.code_build_run_name,
            puppet_account_id=self.puppet_account_id,
            region=self.region,
            account_id=self.account_id,
            ssm_param_inputs=self.ssm_param_inputs,
            launch_parameters=self.launch_parameters,
            manifest_parameters=self.manifest_parameters,
            account_parameters=self.account_parameters,
            project_name=self.project_name,
            requested_priority=self.requested_priority,
        )
        self.write_output(self.params_for_results_display())


class DoExecuteCodeBuildRunTask(
    CodeBuildRunBaseTask, manifest_tasks.ManifestMixen, dependency.DependenciesMixin
):
    code_build_run_name = luigi.Parameter()
    puppet_account_id = luigi.Parameter()

    region = luigi.Parameter()
    account_id = luigi.Parameter()

    ssm_param_inputs = luigi.ListParameter(default=[], significant=False)

    launch_parameters = luigi.DictParameter(default={}, significant=False)
    manifest_parameters = luigi.DictParameter(default={}, significant=False)
    account_parameters = luigi.DictParameter(default={}, significant=False)

    project_name = luigi.Parameter()
    requested_priority = luigi.IntParameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "code_build_run_name": self.code_build_run_name,
            "region": self.region,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def api_calls_used(self):
        return [
            f"codebuild.start_build_{self.puppet_account_id}_{self.project_name}",
            f"codebuild.batch_get_projects_{self.puppet_account_id}_{self.project_name}",
        ]

    def requires(self):
        requirements = {
            "ssm_params": self.get_ssm_parameters(),
        }
        return requirements

    def run(self):
        with self.hub_client("codebuild") as codebuild:
            provided_parameters = self.get_parameter_values()
            parameters_to_use = list()

            projects = codebuild.batch_get_projects(names=[self.project_name]).get(
                "projects", []
            )
            for project in projects:
                if project.get("name") == self.project_name:
                    for environment_variable in project.get("environment", {}).get(
                        "environmentVariables", []
                    ):
                        if environment_variable.get("type") == "PLAINTEXT":
                            n = environment_variable.get("name")
                            if provided_parameters.get(n):
                                parameters_to_use.append(
                                    dict(
                                        name=n,
                                        value=provided_parameters.get(n),
                                        type="PLAINTEXT",
                                    )
                                )

            parameters_to_use.append(
                dict(name="TARGET_ACCOUNT_ID", value=self.account_id, type="PLAINTEXT",)
            )
            parameters_to_use.append(
                dict(name="TARGET_REGION", value=self.region, type="PLAINTEXT",)
            )
            codebuild.start_build_and_wait_for_completion(
                projectName=self.project_name,
                environmentVariablesOverride=parameters_to_use,
            )
        self.write_output(self.params_for_results_display())


class CodeBuildRunForTask(CodeBuildRunBaseTask, manifest_tasks.ManifestMixen):
    code_build_run_name = luigi.Parameter()
    puppet_account_id = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "code_build_run_name": self.code_build_run_name,
            "cache_invalidator": self.cache_invalidator,
        }

    def get_klass_for_provisioning(self):
        return ExecuteCodeBuildRunTask

    def run(self):
        self.write_output(self.params_for_results_display())


class CodeBuildRunForRegionTask(CodeBuildRunForTask):
    region = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "code_build_run_name": self.code_build_run_name,
            "region": self.region,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        dependencies = list()
        these_dependencies = list()
        requirements = dict(
            dependencies=dependencies, these_dependencies=these_dependencies,
        )

        klass = self.get_klass_for_provisioning()

        for task in self.manifest.get_tasks_for_launch_and_region(
            self.puppet_account_id,
            self.section_name,
            self.code_build_run_name,
            self.region,
        ):
            dependencies.append(
                klass(**task, manifest_file_path=self.manifest_file_path)
            )

        return requirements


class CodeBuildRunForAccountTask(CodeBuildRunForTask):
    account_id = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "code_build_run_name": self.code_build_run_name,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        dependencies = list()
        requirements = dict(dependencies=dependencies,)

        klass = self.get_klass_for_provisioning()

        for task in self.manifest.get_tasks_for_launch_and_account(
            self.puppet_account_id,
            self.section_name,
            self.code_build_run_name,
            self.account_id,
        ):
            dependencies.append(
                klass(**task, manifest_file_path=self.manifest_file_path)
            )

        return requirements


class CodeBuildRunForAccountAndRegionTask(CodeBuildRunForTask):
    account_id = luigi.Parameter()
    region = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "code_build_run_name": self.code_build_run_name,
            "region": self.region,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        dependencies = list()
        requirements = dict(dependencies=dependencies)

        klass = self.get_klass_for_provisioning()

        for task in self.manifest.get_tasks_for_launch_and_account_and_region(
            self.puppet_account_id,
            self.section_name,
            self.code_build_run_name,
            self.account_id,
            self.region,
        ):
            dependencies.append(
                klass(**task, manifest_file_path=self.manifest_file_path)
            )

        return requirements


class CodeBuildRunTask(CodeBuildRunForTask):
    code_build_run_name = luigi.Parameter()
    puppet_account_id = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "code_build_run_name": self.code_build_run_name,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        requirements = list()

        klass = self.get_klass_for_provisioning()
        for (
            account_id,
            regions,
        ) in self.manifest.get_account_ids_and_regions_used_for_section_item(
            self.puppet_account_id, self.section_name, self.code_build_run_name
        ).items():
            for region in regions:
                for task in self.manifest.get_tasks_for_launch_and_account_and_region(
                    self.puppet_account_id,
                    self.section_name,
                    self.code_build_run_name,
                    account_id,
                    region,
                ):
                    requirements.append(
                        klass(**task, manifest_file_path=self.manifest_file_path)
                    )

        return requirements

    def run(self):
        self.write_output(self.params_for_results_display())


class CodeBuildRunsSectionTask(CodeBuildRunBaseTask, manifest_tasks.SectionTask):
    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        requirements = list()

        for name, details in self.manifest.get(constants.CODE_BUILD_RUNS, {}).items():
            requirements += self.handle_requirements_for(
                name,
                details,
                constants.CODE_BUILD_RUN,
                constants.CODE_BUILD_RUNS,
                self.execution_mode == constants.EXECUTION_MODE_SPOKE,
                CodeBuildRunForRegionTask,
                CodeBuildRunForAccountTask,
                CodeBuildRunForAccountAndRegionTask,
                CodeBuildRunTask,
                dict(
                    code_build_run_name=name,
                    puppet_account_id=self.puppet_account_id,
                    manifest_file_path=self.manifest_file_path,
                ),
            )

        return requirements

    def run(self):
        self.write_output(self.manifest.get(self.section_name))



from servicecatalog_puppet import constants


def generate_dependency_tasks(
    dependencies, manifest_file_path, puppet_account_id, account_id, region
):
    from servicecatalog_puppet.workflow import codebuild_runs
    from servicecatalog_puppet.workflow import launch
    from servicecatalog_puppet.workflow import spoke_local_portfolios
    from servicecatalog_puppet.workflow import assertions
    from servicecatalog_puppet.workflow import lambda_invocations

    these_dependencies = list()
    common_args = dict(
        manifest_file_path=manifest_file_path, puppet_account_id=puppet_account_id,
    )
    for depends_on in dependencies:
        if depends_on.get("type") == constants.LAUNCH:
            if depends_on.get(constants.AFFINITY) == constants.LAUNCH:
                these_dependencies.append(
                    launch.LaunchTask(
                        **common_args, launch_name=depends_on.get("name"),
                    )
                )
            if depends_on.get(constants.AFFINITY) == "account":
                these_dependencies.append(
                    launch.LaunchForAccountTask(
                        **common_args,
                        launch_name=depends_on.get("name"),
                        account_id=account_id,
                    )
                )
            if depends_on.get(constants.AFFINITY) == "region":
                these_dependencies.append(
                    launch.LaunchForRegionTask(
                        **common_args,
                        launch_name=depends_on.get("name"),
                        region=region,
                    )
                )
            if depends_on.get(constants.AFFINITY) == "account-and-region":
                these_dependencies.append(
                    launch.LaunchForAccountAndRegionTask(
                        **common_args,
                        launch_name=depends_on.get("name"),
                        account_id=account_id,
                        region=region,
                    )
                )

        elif depends_on.get("type") == constants.SPOKE_LOCAL_PORTFOLIO:
            if depends_on.get(constants.AFFINITY) == constants.SPOKE_LOCAL_PORTFOLIO:
                these_dependencies.append(
                    spoke_local_portfolios.SpokeLocalPortfolioTask(
                        **common_args,
                        spoke_local_portfolio_name=depends_on.get("name"),
                    )
                )
            if depends_on.get(constants.AFFINITY) == "account":
                these_dependencies.append(
                    spoke_local_portfolios.SpokeLocalPortfolioForAccountTask(
                        **common_args,
                        spoke_local_portfolio_name=depends_on.get("name"),
                        account_id=account_id,
                    )
                )
            if depends_on.get(constants.AFFINITY) == "region":
                these_dependencies.append(
                    spoke_local_portfolios.SpokeLocalPortfolioForRegionTask(
                        **common_args,
                        spoke_local_portfolio_name=depends_on.get("name"),
                        region=region,
                    )
                )
            if depends_on.get(constants.AFFINITY) == "account-and-region":
                these_dependencies.append(
                    spoke_local_portfolios.SpokeLocalPortfolioForAccountAndRegionTask(
                        **common_args,
                        spoke_local_portfolio_name=depends_on.get("name"),
                        account_id=account_id,
                        region=region,
                    )
                )

        elif depends_on.get("type") == constants.ASSERTION:
            if depends_on.get(constants.AFFINITY) == constants.ASSERTION:
                these_dependencies.append(
                    assertions.AssertionTask(
                        **common_args, assertion_name=depends_on.get("name"),
                    )
                )
            if depends_on.get(constants.AFFINITY) == "account":
                these_dependencies.append(
                    assertions.AssertionForAccountTask(
                        **common_args,
                        assertion_name=depends_on.get("name"),
                        account_id=account_id,
                    )
                )
            if depends_on.get(constants.AFFINITY) == "region":
                these_dependencies.append(
                    assertions.AssertionForRegionTask(
                        **common_args,
                        assertion_name=depends_on.get("name"),
                        region=region,
                    )
                )
            if depends_on.get(constants.AFFINITY) == "account-and-region":
                these_dependencies.append(
                    assertions.AssertionForAccountAndRegionTask(
                        **common_args,
                        assertion_name=depends_on.get("name"),
                        account_id=account_id,
                        region=region,
                    )
                )

        elif depends_on.get("type") == constants.CODE_BUILD_RUN:
            if depends_on.get(constants.AFFINITY) == constants.CODE_BUILD_RUN:
                these_dependencies.append(
                    codebuild_runs.CodeBuildRunTask(
                        **common_args, code_build_run_name=depends_on.get("name"),
                    )
                )
            if depends_on.get(constants.AFFINITY) == "account":
                these_dependencies.append(
                    codebuild_runs.CodeBuildRunForAccountTask(
                        **common_args,
                        code_build_run_name=depends_on.get("name"),
                        account_id=account_id,
                    )
                )
            if depends_on.get(constants.AFFINITY) == "region":
                these_dependencies.append(
                    codebuild_runs.CodeBuildRunForRegionTask(
                        **common_args,
                        code_build_run_name=depends_on.get("name"),
                        region=region,
                    )
                )
            if depends_on.get(constants.AFFINITY) == "account-and-region":
                these_dependencies.append(
                    codebuild_runs.CodeBuildRunForAccountAndRegionTask(
                        **common_args,
                        code_build_run_name=depends_on.get("name"),
                        account_id=account_id,
                        region=region,
                    )
                )

        elif depends_on.get("type") == constants.LAMBDA_INVOCATION:
            if depends_on.get(constants.AFFINITY) == constants.LAMBDA_INVOCATION:
                these_dependencies.append(
                    lambda_invocations.LambdaInvocationTask(
                        **common_args, lambda_invocation_name=depends_on.get("name"),
                    )
                )
            if depends_on.get(constants.AFFINITY) == "account":
                these_dependencies.append(
                    lambda_invocations.LambdaInvocationForAccountTask(
                        **common_args,
                        lambda_invocation_name=depends_on.get("name"),
                        account_id=account_id,
                    )
                )
            if depends_on.get(constants.AFFINITY) == "region":
                these_dependencies.append(
                    lambda_invocations.LambdaInvocationForRegionTask(
                        **common_args,
                        lambda_invocation_name=depends_on.get("name"),
                        region=region,
                    )
                )
            if depends_on.get(constants.AFFINITY) == "account-and-region":
                these_dependencies.append(
                    lambda_invocations.LambdaInvocationForAccountAndRegionTask(
                        **common_args,
                        lambda_invocation_name=depends_on.get("name"),
                        account_id=account_id,
                        region=region,
                    )
                )
    return these_dependencies


class DependenciesMixin(object):
    def get_section_dependencies(self):
        from servicecatalog_puppet.workflow import codebuild_runs
        from servicecatalog_puppet.workflow import launch
        from servicecatalog_puppet.workflow import spoke_local_portfolios
        from servicecatalog_puppet.workflow import assertions
        from servicecatalog_puppet.workflow import lambda_invocations

        from servicecatalog_puppet.workflow import generate

        if isinstance(self, codebuild_runs.ExecuteCodeBuildRunTask):
            item_name = self.code_build_run_name
        elif isinstance(self, launch.ProvisioningTask):
            item_name = self.launch_name
        elif isinstance(self, spoke_local_portfolios.SpokeLocalPortfolioBaseTask):
            item_name = self.spoke_local_portfolio_name
        elif isinstance(self, assertions.AssertTask):
            item_name = self.assertion_name
        elif isinstance(self, lambda_invocations.InvokeLambdaTask):
            item_name = self.lambda_invocation_name

        dependencies = (
            self.manifest.get(self.section_name).get(item_name).get("depends_on", [])
        )

        these_dependencies = generate_dependency_tasks(
            dependencies,
            self.manifest_file_path,
            self.puppet_account_id,
            self.account_id,
            self.region,
        )

        if self.section_name in [constants.SPOKE_LOCAL_PORTFOLIOS, constants.LAUNCHES]:
            if not (
                self.execution_mode == constants.EXECUTION_MODE_SPOKE or self.is_dry_run
            ):
                these_dependencies.append(
                    generate.GenerateSharesTask(
                        puppet_account_id=self.puppet_account_id,
                        manifest_file_path=self.manifest_file_path,
                        section=self.section_name,
                    )
                )

        return these_dependencies
